
/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */

package com.jsevy.jodg;


import java.awt.geom.AffineTransform;

/**
 * Graphical entity representing a circular arc.
 * 
 * @author jsevy
 *
 */
public class ODGCircle extends ODGEntity
{
    protected RealPoint center;
    protected double radius;
    protected double startAngle;
    protected double sweepAngle;
    protected ODGGraphicsStyle style;
    
    
    /**
     * Create a circular arc from the specified parameters.
     * @param center		Center of the circle, as a RealPoint
     * @param radius		Radius of the circle
     * @param startAngle	Starting angle of the circle, in degrees, counterclockwise from the positive x axis
     * @param sweepAngle	Sweep angle of the circle, in degrees, with positive being counterclockwise from the positive x axis
     * @param style			The ODGSTyle object specifying rendering parameters for the shape (color, thickness, etc.)
     */
    public ODGCircle(RealPoint center, double radius, double startAngle, double sweepAngle, ODGGraphicsStyle style)
    {
        this(center, radius, startAngle, sweepAngle, style, new AffineTransform());
    }
    
    
    /**
     * Create a circular arc from the specified parameters.
     * @param center        Center of the circle, as a RealPoint
     * @param radius        Radius of the circle
     * @param startAngle    Starting angle of the circle, in degrees, counterclockwise from the positive x axis
     * @param sweepAngle    Sweep angle of the circle, in degrees, with positive being counterclockwise from the positive x axis
     * @param style         The ODGSTyle object specifying rendering parameters for the shape (color, thickness, etc.)
     * @param transform     Affine transform to be applied to the shape
     */
    public ODGCircle(RealPoint center, double radius, double startAngle, double sweepAngle, ODGGraphicsStyle style, AffineTransform transform)
    {
        super(transform);
        
        this.startAngle = startAngle;
        this.sweepAngle = sweepAngle;
        this.center = center;
        this.radius = radius;
        this.style = style;
        
        populateXML();
    }
    
    /**
     * Populate XMLNode fields for XML generation
     */
    protected void populateXML()
    {
        // add transform attribute if any
        super.populateXML();
        
        this.setNodeName("draw:circle");
        
        // add attributes
        this.setAttribute("svg:cx", scaleValueWithUnits(center.x));
        this.setAttribute("svg:cy", scaleValueWithUnits(center.y));
        this.setAttribute("svg:r", scaleValueWithUnits(radius));
        
        // only use angles if not full circle/ellipse
        if (sweepAngle < 360)
        {
            // adjust angles since all arcs go counterclockwise
            double endAngle;
            if (sweepAngle > 0)
            {
                endAngle = startAngle + sweepAngle;
            }
            else
            {
                endAngle = startAngle;
                startAngle = endAngle + sweepAngle;
            }
            
            // also all angles must be between -360 and 360
            while (startAngle >= 360)
                startAngle -= 360;
            while (startAngle <= -360)
                startAngle += 360;
            
            while (endAngle >= 360)
                endAngle -= 360;
            while (endAngle <= -360)
                endAngle += 360;
            
            this.setAttribute("draw:start-angle", Double.toString(startAngle));
            this.setAttribute("draw:end-angle", Double.toString(endAngle));
            
            // kind attribute depends on the style - filled or not
            if (style.filled)
            {
                this.setAttribute("draw:kind", "section");
            }
            else
            {
                this.setAttribute("draw:kind", "arc");
            }
            
        }
        else
        {
            // full circle
            this.setAttribute("draw:kind", "full");
        }
        
        
        
        this.setAttribute("draw:style-name", style.getStyleName());
        
    }
    
    
    
}