/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jodg;



import java.awt.Graphics2D;
import java.io.StringWriter;
import java.util.Vector;

import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import com.jsevy.jxml.XMLDocument;
import com.jsevy.jxml.XMLElement;

/**
 * Class representing an Open Document Format drawing document, which owns an odgGraphics on which drawing commands can be made. The document's
 * toODGString method can then be called to generate the ODG text corresponding to the drawing done on the Graphics object.
 * 
 * The typical workflow is as follows:
 * 
 * <pre>
 * 
 * // Create an odg document and get its associated ODGGraphics instance
 * ODGDocument odgDocument = new ODGDocument(); 
 * ODGGraphics odgGraphics = odgDocument.getGraphics(); 
 * 
 * // Do drawing commands as on any other Graphics. If you have a paint(Graphics) method, 
 * // you can just use it with the ODGGraphics instance since it's a subclass of Graphics. 
 * graphics.setColor(Color.RED);  
 * graphics.setStroke(new BasicStroke(3));
 * graphics.drawLine(0, 0, 1000, 500); 
 * graphics.drawRect(1000, 500, 150, 150); 
 * graphics.drawRoundRect(20, 200, 130, 100, 20, 10); 
 * 
 * // Get the flat ODG output as a string - it's just XML text - and  save  in a file for use 
 * // with LibreOffice Draw or any other draw package that accepts flat ODG (.fodg) files
 * String stringOutput = odgDocument.toODGString(); 
 * String filePath = "path/to/file.fodg"; 
 * FileWriter fileWriter = new FileWriter(filePath); 
 * fileWriter.write(odgText); 
 * fileWriter.flush(); 
 * fileWriter.close();
 * 
 * </pre>
 * 
 * @author jsevy
 *
 */
public class ODGDocument extends XMLDocument
{
    
    private Vector<ODGEntity> entities;
    
    private ODGGraphics graphics;
    
    private ODGFontFaces fontFaces;
    
    private ODGStyles styles;
    
    private ODGBody body;
    
    private ODGDrawing drawing;
    
    private ODGPage currentPage;
    
    
    
    /**
     * Create a new ODGDocument
     */
    public ODGDocument()
    {
        this.graphics = new ODGGraphics(this);
        this.entities = new Vector<ODGEntity>();
        this.fontFaces = new ODGFontFaces();
        this.styles = new ODGStyles();
        this.body = new ODGBody();
        this.drawing = new ODGDrawing();
        this.currentPage = new ODGPage();
        
        this.getDocumentElement().appendChild(fontFaces);
        this.getDocumentElement().appendChild(styles);
        this.getDocumentElement().appendChild(body);
        body.appendChild(drawing);
        drawing.appendChild(currentPage);
        
        populateXML();
    }
    
    
    
    
    /** Set the units for interpreting the dimension values in the ODG file.
     *
     *  @param unitsCode  ODG code indicating units:
     *      0 = Unitless
     *      1 = Inches
     *      2 = Feet
     *      3 = Miles
     *      4 = Millimeter
     *      5 = Centimeters
     *      6 = Meters
     *      7 = Kilometers
     *      8 = Microinches
     *      9 = Mils
     *      10 = Yards
     * 
     */
    public void setUnits(int unitsCode)
    {
        String[] unitsMatrix = 
            {
            /* 0 = */ "",
            /* 1 = */ "in",
            /* 2 = */ "ft",
            /* 3 = */ "mi",
            /* 4 = */ "mm",
            /* 5 = */ "cm",
            /* 6 = */ "m",
            /* 7 = */ "km",
            /* 8 = */ "uin",
            /* 9 = */ "mils",
            /* 10 = */ "yd",
        };
        
        ODGEntity.units = unitsMatrix[unitsCode];
    }
    
    
    
    /**
     * Set the scale factor for dimensions in the Java drawing - how many Java units/pixels
     * correspong to one unit in the ODG output. Note that the units can be set with a call to 
     * setUnits
     * 
     * @param javaUnitsPerUnit  How many Java units/pixels correspond to one ODG unit
     */
    public void setScaleFactor(double javaUnitsPerUnit)
    {
        ODGEntity.scaleFactor = javaUnitsPerUnit;
    }
    
    
    
    /**
     * When set to true, the current Java transform will be applied to a shape before it is converted into
     * an equivalent ODG object. The default is false; in this case the current Java transform is just embedded 
     * as part of the ODG shape's definition, as discussed in the ODG specification. However, some versions of
     * vector drawing packages, most notably LibreOffice Draw, do not handle certain types of transforms correctly,
     * and thus will incorrectly render some shapes with embedded transforms. By setting this flag to true, the Java
     * transform is used to transform shapes before the ODG is generated, so they won't (for the most part) have
     * embedded transforms. (The exception concerns ellipses - when the major and minor axes aren't parallel to the
     * x and y axes, these need to have a rotation and translation associated since the ODG ellipse definition only 
     * supports horizontal and vertical major and minor axes. Fortunately LibreOffice handles these transforms correctly.)
     * See:
     *      https://bugs.documentfoundation.org/show_bug.cgi?id=98584
     *      https://bugs.documentfoundation.org/show_bug.cgi?id=98565
     * 
     * @param preApplyTransform  When false (the default), any Java transform is embedded in the ODG definition.
     *                           When true, the current Java transform is applied to each shape before its ODG equivalent is generated.
     *                           
     */
    public void preapplyTransform(boolean preApplyTransform)
    {
        ODGEntity.preApplyTransform = preApplyTransform;
    }
    
    
    /**
     * Since the fonts used in an ODG program (e.g., LibreOffice Draw) aren't identical to those used
     * in Java, the physical widths of strings can differ slightly. When the ODG fonts are slightly
     * wider, the text generated by the JODG library may not fit into the text frame created for them
     * when they are displayed in the ODG viewer and may be split into two lines. This factor causes 
     * the text frame width to be increased slightly to accommodate this discrepancy. The default value
     * is set to 1.05, so that frames will be 5% larger than needed. Note that this won't affect the display
     * since the frames are invisible in the ODG drawing. You'll only notice if you go to edit the text.
     * 
     * @param textFrameExpansionFactor  Factor to multiply each text frame's width by (default is 1.05)
     */
    public void setTextFrameExpansionFactor(double textFrameExpansionFactor)
    {
        ODGFrame.expansionFactor = textFrameExpansionFactor;
    }
    
    
    
    /**
     * Get the ODGGraphics associated with this document for use with standard Graphics drawing operations to generate
     * an odg text representation.
     * 
     * @return	The ODGGraphics associated with this document, on which standard Java Graphics drawing calls can be made
     */
    public ODGGraphics getGraphics()
    {
        return this.graphics;
    }
    
    
    /**
     * Get a graphics style corresponding to the current Graphics2D instance, and whether the shape
     * should be stroked and/or filled. Will return an existing corresponding style if there is one,
     * or create a new one and add it to the list of global styles if needed.
     * 
     * @param graphics  Java Graphics2D instance
     * @param stroked   True if shape is to be stroked with current color of supplied graphics instance
     * @param filled    True if shape is to be filled with current color of supplied graphics instance
     * @return          Corresponding ODGGraphicsStyle
     */
    protected ODGGraphicsStyle getGraphicsStyle(Graphics2D graphics, boolean stroked, boolean filled)
    {
        return styles.getGraphicsStyle(graphics, stroked, filled);
    }
    
    
    /**
     * Get a text style corresponding to the current Graphics2D instance. Will return an existing corresponding
     * style if there is one, or create a new one and add it to the list of global styles if needed.
     * 
     * @param graphics  Java Graphics2D instance
     * @return          Corresponding ODGTextStyle
     */
    protected ODGTextStyle getTextStyle(Graphics2D graphics)
    {
        return styles.getTextStyle(graphics);
    }
    
    
    private void populateXML()
    {
        ((XMLElement)this.getDocumentElement()).setNodeName("office:document");
        
        // add namespace attributes - whew!
        this.getDocumentElement().setAttribute("xmlns:officeooo", "http://openoffice.org/2009/office");
        this.getDocumentElement().setAttribute("xmlns:anim", "urn:oasis:names:tc:opendocument:xmlns:animation:1.0");
        this.getDocumentElement().setAttribute("xmlns:smil", "urn:oasis:names:tc:opendocument:xmlns:smil-compatible:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:presentation", "urn:oasis:names:tc:opendocument:xmlns:presentation:1.0");
        this.getDocumentElement().setAttribute("xmlns:grddl", "http://www.w3.org/2003/g/data-view#");
        this.getDocumentElement().setAttribute("xmlns:xhtml", "http://www.w3.org/1999/xhtml");
        this.getDocumentElement().setAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance");
        this.getDocumentElement().setAttribute("xmlns:xsd", "http://www.w3.org/2001/XMLSchema");
        this.getDocumentElement().setAttribute("xmlns:xforms", "http://www.w3.org/2002/xforms");
        this.getDocumentElement().setAttribute("xmlns:dom", "http://www.w3.org/2001/xml-events");
        this.getDocumentElement().setAttribute("xmlns:script", "urn:oasis:names:tc:opendocument:xmlns:script:1.0" );
        this.getDocumentElement().setAttribute("xmlns:form", "urn:oasis:names:tc:opendocument:xmlns:form:1.0" );
        this.getDocumentElement().setAttribute("xmlns:math", "http://www.w3.org/1998/Math/MathML");
        this.getDocumentElement().setAttribute("xmlns:draw", "urn:oasis:names:tc:opendocument:xmlns:drawing:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:dr3d", "urn:oasis:names:tc:opendocument:xmlns:dr3d:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:text", "urn:oasis:names:tc:opendocument:xmlns:text:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:style", "urn:oasis:names:tc:opendocument:xmlns:style:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:formx", "urn:openoffice:names:experimental:ooxml-odf-interop:xmlns:form:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:dc", "http://purl.org/dc/elements/1.1/");
        this.getDocumentElement().setAttribute("xmlns:ooo", "http://openoffice.org/2004/office"); 
        this.getDocumentElement().setAttribute("xmlns:loext", "urn:org:documentfoundation:names:experimental:office:xmlns:loext:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:office", "urn:oasis:names:tc:opendocument:xmlns:office:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:fo", "urn:oasis:names:tc:opendocument:xmlns:xsl-fo-compatible:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:field", "urn:openoffice:names:experimental:ooo-ms-interop:xmlns:field:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:xlink", "http://www.w3.org/1999/xlink"); 
        this.getDocumentElement().setAttribute("xmlns:meta", "urn:oasis:names:tc:opendocument:xmlns:meta:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:config", "urn:oasis:names:tc:opendocument:xmlns:config:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:calcext", "urn:org:documentfoundation:names:experimental:calc:xmlns:calcext:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:svg", "urn:oasis:names:tc:opendocument:xmlns:svg-compatible:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:of", "urn:oasis:names:tc:opendocument:xmlns:of:1.2"); 
        this.getDocumentElement().setAttribute("xmlns:chart", "urn:oasis:names:tc:opendocument:xmlns:chart:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:rpt", "http://openoffice.org/2005/report"); 
        this.getDocumentElement().setAttribute("xmlns:table", "urn:oasis:names:tc:opendocument:xmlns:table:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:css3t", "http://www.w3.org/TR/css3-text/"); 
        this.getDocumentElement().setAttribute("xmlns:number", "urn:oasis:names:tc:opendocument:xmlns:datastyle:1.0"); 
        this.getDocumentElement().setAttribute("xmlns:ooow", "http://openoffice.org/2004/writer"); 
        this.getDocumentElement().setAttribute("xmlns:oooc", "http://openoffice.org/2004/calc"); 
        this.getDocumentElement().setAttribute("xmlns:tableooo", "http://openoffice.org/2009/table"); 
        this.getDocumentElement().setAttribute("xmlns:drawooo", "http://openoffice.org/2010/draw"); 
        this.getDocumentElement().setAttribute("office:version", "1.2"); 
        this.getDocumentElement().setAttribute("office:mimetype", "application/vnd.oasis.opendocument.graphics");

    }
    
    
    
    /**
     * Return the ODG text associated with this ODG document. This includes the header,
     * classes, tables, blocks, entities and objects sections, populated with content
     * generated by graphics calls on the associated ODGGraphics.
     * 
     * @return The ODG text associated with this document.
     */
    public String toODGString()
    {
        
        DOMSource domSource = new DOMSource(this);
        StringWriter stringWriter = new StringWriter();
        
        try
        {
            TransformerFactory tf = TransformerFactory.newInstance();
            Transformer transformer = tf.newTransformer();
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "no");
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            
            transformer.transform(domSource, new StreamResult(stringWriter));
        }
        catch (TransformerException e)
        {
            // really shouldn't ever happen...
            System.out.println("TransformerException while creating ODG XML: " + e.getMessage());
        }
        
        //xmlBuffer.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + "\n");
        //xmlBuffer.append(this.toXMLString(0, 2));
        
        StringBuffer xmlBuffer = stringWriter.getBuffer();
        
        // Well this is annoying - for text, the whitespace in pretty-printed XML is interpreted in LibreOffice Draw
        // as (one) extra space appended to the text, which comes in at the default size set in LibreOffice Draw, 
        // which then mis-aligns the rest of the text if the default is a larger font size. What's needed is
        // just for there to be no space/tabs/returns between the </text:span> and </text:p> tags, but
        // I don't know how to get the Transformer to do only that while pretty-printing the rest. So we
        // post-process the XML string to remove the spaces... Perhaps an appropriate XSLT specification could
        // force the transformer to do this, but this just seemed easier (if hack-y...)   ¯\_(ツ)_/¯
        xmlBuffer = removeSpacesBetweenStrings(xmlBuffer, "<text:p>", "<text:span");
        xmlBuffer = removeSpacesBetweenStrings(xmlBuffer, "</text:span>", "</text:p>");
        
        return xmlBuffer.toString();
    }
    
    /**
     * Remove any whitespace characters between any two successive occurrences of startTag and endTag. Used
     * here just to remove pretty-printed white spaces between <text:p> and <text:span> and their closing
     * tags.
     * @param buf       StringBuffer containing the string to be searched
     * @param startTag  Beginning tag
     * @param endTag    Ending tag
     * @return          The modified StringBuffer
     */
    private StringBuffer removeSpacesBetweenStrings(StringBuffer buf, String startTag, String endTag)
    {
        int startIndex = 0;
        int parentLength = startTag.length();
        
        startIndex = buf.indexOf(startTag, startIndex);
        while (startIndex > 0)
        {
            // move to end of start tag
            startIndex += parentLength;
            
            // find following child
            int endIndex = buf.indexOf(endTag, startIndex);
            if (endIndex > 0)
            {
                // strip whitespace between them
                while (startIndex < endIndex)
                {
                    if (Character.isWhitespace(buf.charAt(startIndex)))
                    {
                        // remove character, and decrement endIndex since character gone
                        buf.deleteCharAt(startIndex);
                        endIndex--;
                    }
                    else
                    {
                        // increment index
                        startIndex++;
                    }
                }
                
                // find next start tag
                startIndex = buf.indexOf(startTag, startIndex);
            }
        }
        
        return buf;
    }
    
    
    /**
     * Utility method used by the associated ODGGraphics object.
     * 
     * @param entity an odgEntity instance
     */
    public void addEntity(ODGEntity entity)
    {
        entities.add(entity);
        currentPage.appendChild(entity);
    }
    
}