/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jodg;


import java.awt.geom.AffineTransform;



/**
 * A special case of DXFObjects used for graphical entities. In particular, maintains the affine transformation
 * that can be associated with any graphical element.
 * 
 * @author jsevy
 *
 */
public class ODGEntity extends ODGObject
{
    protected static boolean preApplyTransform = false;
    
    // transform for this object if any
    protected AffineTransform transform = new AffineTransform();
    
    private double translateX, translateY;
    private double rotateAngleRadians;
    private double scaleX, scaleY;
    private double skewXAngleRadians;
    
    
    
    public ODGEntity(AffineTransform transform)
    {
        super();
        this.transform = transform;
        
        decomposeTransform();
    }
    
    
    public AffineTransform getTransform()
    {
        return transform;
    }
    
    
    public void setTransform(AffineTransform newTransform)
    {
        this.transform = newTransform;
    }
    
    
    
    /**
     * Method called by subclasses to populate the XML with information regarding the affine transformation
     * (omitted if it's the identity)
     */
    protected void populateXML()
    {
        // add attribute for transform, if needed
        if (!transform.isIdentity())
        {
            String transformCoeffs = "";
            double coeffs[] = new double[6];
            transform.getMatrix(coeffs);
            transformCoeffs = "matrix(" + coeffs[0] + " "
                                         + coeffs[1] + " "
                                         + coeffs[2] + " "
                                         + coeffs[3] + " "
                                         + scaleValueWithUnits(coeffs[4]) + " "
                                         + scaleValueWithUnits(coeffs[5]) + ")";
            this.setAttribute("draw:transform", transformCoeffs);
        }
        
        
    }
    
    
    
    
    protected void newpopulateXML()
    {
        // add attribute for transform, if needed
        if (!transform.isIdentity())
        {
            String transformString = new String();
            
            if ((translateX != 0) || (translateY != 0))
                transformString += "translate(" + scaleValueWithUnits(translateX) + " " + scaleValueWithUnits(translateY) + ")";
            
            if (rotateAngleRadians != 0)
                transformString += " rotate(" + rotateAngleRadians + ")";
            
            if ((scaleX != 1) || (scaleY != 1))
                transformString += " scale(" + scaleX + " " + scaleY + ")";
            
            if (skewXAngleRadians != 0)
                transformString += " skewX(" + skewXAngleRadians + ")";
            
            this.setAttribute("draw:transform", transformString);
        }
        
        
    }
    
    private void decomposeTransform()
    {
        // well this sucks... but the general matrix method doesn't seem to work
        // using stuff from: http://frederic-wang.fr/decomposition-of-2d-transform-matrices.html
        // decompose matrix 
        //          [a, c, e]
        //          [b, d, f]
        //          [0, 0, 1]
        // into
        //   translate(e,f) rotate(sign(b)*acos(a/r)) scale(r, Δ/r) skewX(atan((ac+bd)/r^2))
        // where
        //   Δ = ad - bc    and   r = sqrt(a^2 + b^2)
        //
        // flat matrix array is  [a, b, c, d, e, f]
        
        double coeffs[] = new double[6];
        transform.getMatrix(coeffs);
        double a = coeffs[0];
        double b = coeffs[1];
        double c = coeffs[2];
        double d = coeffs[3];
        double e = coeffs[4];
        double f = coeffs[5];
        
        double det = a*d - b*c;
        double r = Math.sqrt(a*a + b*b);
        
        translateX = e;
        translateY = f;
        rotateAngleRadians = Math.signum(b) * Math.acos(a/r);
        scaleX = r;
        scaleY = det/r;
        skewXAngleRadians = Math.atan((a*c + b*d)/(r*r));
        
    }
    
    
    
}