/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */

package com.jsevy.jodg;


import java.awt.FontMetrics;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;



/**
 * Class representing a container for ODG text, located at a specified position and having specified style (font, size).
 * 
 * @author jsevy
 *
 */
public class ODGFrame extends ODGEntity
{
    private RealPoint topPoint;
    private double width, height;
    private ODGTextBox textBox;
    private ODGGraphicsStyle style;
    
    // factor to increase the width of text frames when generated, so that strings won't be
    // split into two lines when the ODG font dimensions are slightly different than the
    // corresponding Java font dimensions. The value can be set from the method setTextFrameExpansionFactor
    // within the ODGDoument class.
    protected static double expansionFactor = 1.05;
    
    
    /**
     * Create a frame containing text at the specified position.
     * 
     * @param text		        Text string
     * @param basePoint	        Location of the base point of the text
     * @param graphicsStyle		Style providing graphics rendering info (color, etc)
     * @param textStyle         Style providing text rendering info (font, etc.)
     * @param graphics          The Java graphics object, used to get some font metrics info needed to size the frame
     * @param transform         Affine transform to be applied to the frame
     */
    public ODGFrame(String text, RealPoint basePoint, ODGGraphicsStyle graphicsStyle, ODGTextStyle textStyle, Graphics2D graphics, AffineTransform transform)
    {
        super(transform);
        
        // need dimensions of box, and adjustment for top point of frame; FontMetrics will get these for us
        FontMetrics fm = graphics.getFontMetrics();
        int ascent = fm.getAscent();
        int descent = fm.getDescent();
        int stringWidth = fm.stringWidth(text);
        
        // top point used rather than text line base point; need to add ascent to baseline
        this.topPoint = new RealPoint(basePoint);
        topPoint.y -= ascent;
                
        this.width = expansionFactor * stringWidth;
        this.height = ascent + descent;
        
        this.style = graphicsStyle;
        this.textBox = new ODGTextBox(text, textStyle, transform);
        
        populateXML();
    }
    
    
    /**
     * Create a frame containing text at the specified position.
     * 
     * @param text              Text string
     * @param basePoint         Location of the base point of the text
     * @param graphicsStyle     Style providing graphics rendering info (color, etc)
     * @param textStyle         Style providing text rendering info (font, etc.)
     * @param graphics          The Java graphics object, used to get some font metrics info needed to size the frame
     */
    public ODGFrame(String text, RealPoint basePoint, ODGGraphicsStyle style, ODGTextStyle textStyle, Graphics2D graphics)
    {
        this(text, basePoint, style, textStyle, graphics, new AffineTransform());
    }
    
    /**
     * Populate XMLNode fields for XML generation
     */
    protected void populateXML()
    {
        // add transform attribute if any
        super.populateXML();
        
        this.setNodeName("draw:frame");
        
        // add attributes - style will make box invisible (no stroke or fill)
        this.setAttribute("draw:style-name", style.getStyleName());
        
        this.setAttribute("svg:x", scaleValueWithUnits(topPoint.x));
        this.setAttribute("svg:y", scaleValueWithUnits(topPoint.y));
        this.setAttribute("svg:width", scaleValueWithUnits(width));
        this.setAttribute("svg:height", scaleValueWithUnits(height));
        
        // now add text-box element as child
        this.appendChild(textBox);
                
    }
    
    
}