/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */



package com.jsevy.jodg;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;

import com.jsevy.jxml.XMLElement;


/**
 * Class for style elements that apply to graphics rendering for things like stroke characteristics and color.
 * @author jsevy
 *
 */
public class ODGGraphicsStyle extends ODGStyle
{

    protected boolean stroked;
    protected boolean filled;
    protected BasicStroke stroke;
    protected Color color;
    
    
    /**
     * Create a new style element for graphics operations.
     * 
     * @param graphics  The Java graphics object, used to get stroke and color info
     * @param stroked   True if the object is to be outlined with the current color
     * @param filled    True if the object is to be filled with the current color
     */
    public ODGGraphicsStyle(Graphics2D graphics, boolean stroked, boolean filled)
    {
        // base class gets style name
        
        this.stroked = stroked;
        this.filled = filled;
        this.color = graphics.getColor();
        
        this.stroke = (BasicStroke)graphics.getStroke();
        
        populateXML();
    }
    
    
    /**
     * Used to determine if a given style is already in the global list of styles so that we
     * don't create redundant identical styles.
     */
    public boolean equals(Object other)
    {
        if (other instanceof ODGGraphicsStyle)
        {
            ODGGraphicsStyle otherGraphics = (ODGGraphicsStyle)other;
            if (    (stroked == otherGraphics.stroked) 
                 && (filled == otherGraphics.filled) 
                 && (color.equals(otherGraphics.color)) 
                 && (stroke.equals(otherGraphics.stroke)) 
               )
                return true;
            else
                return false;
        }
        else
            return false;
    }
    
    
    /**
     * Populate XMLNode fields for XML generation
     */
    private void populateXML()
    {
        // done by base class
        //this.setName("style:style");
        
        // add attributes
        this.setAttribute("style:name", styleName);
        this.setAttribute("style:family", "graphic");
        this.setAttribute("style:parent-style-name", "standard");
        
        
        // add child for style specification
        XMLElement graphicsProperties = new XMLElement();
        graphicsProperties.setNodeName("style:graphic-properties");
        
        // filled?
        if(filled)
        {
            graphicsProperties.setAttribute("draw:fill", "solid");
            
            // set fill color
            graphicsProperties.setAttribute("draw:fill-color", "#" + (Integer.toHexString(color.getRGB()).substring(2)));

        }
        else
        {
            graphicsProperties.setAttribute("draw:fill", "none");
        }
        
        // stroked?
        if(stroked)
        {
            graphicsProperties.setAttribute("draw:stroke", "solid");
            
            // set line width
            graphicsProperties.setAttribute("svg:stroke-width", scaleValueWithUnits(stroke.getLineWidth()));
            
            // set line color
            graphicsProperties.setAttribute("svg:stroke-color", "#" + (Integer.toHexString(color.getRGB()).substring(2)));

        }
        else
        {
            graphicsProperties.setAttribute("draw:stroke", "none");
        }
        
        // make padding 0 for text in text boxes
        graphicsProperties.setAttribute("fo:padding-top", "0");
        graphicsProperties.setAttribute("fo:padding-bottom", "0"); 
        graphicsProperties.setAttribute("fo:padding-left", "0"); 
        graphicsProperties.setAttribute("fo:padding-right", "0");
        
        
        this.appendChild(graphicsProperties);
        
    }
    
}
