/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jodg;


import java.awt.geom.AffineTransform;




/**
 * Class representing a line segment.
 * @author jsevy
 *
 */
public class ODGLine extends ODGEntity
{
    private RealPoint start, end;
    private ODGGraphicsStyle style;
    
    
    /**
     * Create a line segment between the specified endpoints.
     * 
     * @param start		One endpoint
     * @param end		The other endpoint
     * @param style         The ODGSTyle object specifying rendering parameters for the shape (color, thickness, etc.)
     * @param transform     Affine transform to be applied to the shape
     */
    public ODGLine(RealPoint start, RealPoint end, ODGGraphicsStyle style, AffineTransform transform)
    {
        // so, LibreOffice sucks at line transformations - ignores scaling and screws up shear
        // so just transform the end points directly and screw passing along the transform
        //super(transform);
        super(new AffineTransform());
        
        double[] coords = {start.x, start.y, end.x, end.y};
        transform.transform(coords, 0, coords, 0, 2);
        
        this.start = new RealPoint(coords[0], coords[1], 0);
        this.end = new RealPoint(coords[2], coords[3], 0);
        this.style = style;
        
        populateXML();
    }
    
    
    /**
     * Create a line segment between the specified endpoints.
     * 
     * @param start     One endpoint
     * @param end       The other endpoint
     * @param style         The ODGSTyle object specifying rendering parameters for the shape (color, thickness, etc.)
     */
    public ODGLine(RealPoint start, RealPoint end, ODGGraphicsStyle style)
    {
        this(start, end, style, new AffineTransform());
    }
    
    
    
    /**
     * Populate XMLNode fields for XML generation
     */
    protected void populateXML()
    {
        // add transform attribute if any, though we've fixed it to the identity because
        // Libreoffice doesn't handle transforms correctly on lines (!)
        super.populateXML();
        
        this.setNodeName("draw:line");
        
        // add attributes
        this.setAttribute("svg:x1", scaleValueWithUnits(start.x));
        this.setAttribute("svg:y1", scaleValueWithUnits(start.y));
        this.setAttribute("svg:x2", scaleValueWithUnits(end.x));
        this.setAttribute("svg:y2", scaleValueWithUnits(end.y));
        
        this.setAttribute("draw:style-name", style.getStyleName());
        
    }
}