/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jodg;


import java.awt.geom.AffineTransform;
import java.util.Vector;



/**
 * Class representing a set of line segments defining a (possibly closed) polygon.
 * 
 * @author jsevy
 *
 */
public class ODGPolyline extends ODGEntity
{
    private Vector<RealPoint> vertices;
    private ODGGraphicsStyle style;
    
    
    
    /**
     * Create a set of line segments that connects the specified points.
     * 
     * @param vertices		The vertices
     * @param style         The ODGStyle object specifying rendering parameters for the shape (color, thickness, etc.)
     * @param transform     Affine transform to be applied to the shape
     */
    public ODGPolyline(Vector<RealPoint> vertices, ODGGraphicsStyle style, AffineTransform transform)
    {
        super(transform);
        this.style = style;
        this.vertices = vertices;
        
        populateXML();
    }
    
    /**
     * Populate XMLNode fields for XML generation
     */
    protected void populateXML()
    {
        // add transform attribute if any to XML
        super.populateXML();
        
        this.setNodeName("draw:polyline");
        
        // add attributes
        
        // need coords for the viewport, which needs to be located at the minimum of all 
        // the x and y coords in the polyline and has size given by the max x's and y's
        double minX, maxX;
        double minY, maxY;
        
        if (vertices.size() > 0)
        {
            RealPoint point = vertices.elementAt(0);
            
            minX = point.x;
            minY = point.y;
            maxX = point.x;
            maxY = point.y;
        }
        else
        {
            // no points, so arbitrary
            minX = 0;
            minY = 0;
            maxX = 0;
            maxY = 0;
        }
        
        // now add points
        StringBuffer points = new StringBuffer();
        for (int i = 0; i < vertices.size(); i++)
        {
            RealPoint point = vertices.elementAt(i);
            
            points.append(scaleValue(point.x) + "," + scaleValue(point.y) + " ");
            
            // adjust min and max for viewport coords
            minX = Math.min(minX,point.x);
            minY = Math.min(minY,point.y);
            maxX = Math.max(maxX,point.x);
            maxY = Math.max(maxY,point.y);
        }
        
        // set location of viewport containing the polyline; it's the min of all the x and y coords
        this.setAttribute("svg:x", scaleValueWithUnits(minX));
        this.setAttribute("svg:y", scaleValueWithUnits(minY));
        
        // need width and height and viewbox to set scale
        this.setAttribute("svg:width", scaleValueWithUnits(Math.abs(maxX)));
        this.setAttribute("svg:height", scaleValueWithUnits(Math.abs(maxY)));
        this.setAttribute("svg:viewBox", "0 0 " + scaleValue(Math.abs(maxX)) + " " + scaleValue(Math.abs(maxY)));
        
        this.setAttribute("draw:points", points.toString());
        
        this.setAttribute("draw:style-name", style.getStyleName());
        
    }
}