/*
 * JODG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jodg;

import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics2D;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;

import com.jsevy.jxml.XMLElement;




/**
 * Class for style elements that apply to text rendering for things like font characteristics and color.
 * @author jsevy
 *
 */
public class ODGTextStyle extends ODGStyle
{
    
    private Font font;
    private Color color;
    private double transformScaleFactor;
    
    /**
     * Create a new text style.
     * @param graphics  The Java graphics object, used to extract font and color information
     */
    public ODGTextStyle(Graphics2D graphics)
    {
        // base class gets name
        
        this.font = graphics.getFont();
        this.color = graphics.getColor();
        this.transformScaleFactor = getTransformScaleFactor(graphics.getTransform());
        
        populateXML();
    }
    
    
    /**
     * Used to scale Font sizes, since they don't get the transform applied in the resulting ODG.
     * Just take the length that the unit j vector is mapped onto; imperfect since doesn't
     * account for skew, but adequate for transformations with only rotation and (uniform) scaling.
     * @param transform
     * @return
     */
    private double getTransformScaleFactor(AffineTransform transform)
    {
        // see where unit vector j is mapped, and take its length
        // create a version of the transform without the translation to do vector mapping
        double[] matrixCoeffs = {0,0,0,0,0,0};
        transform.getMatrix(matrixCoeffs);
        // zero out translation components
        matrixCoeffs[4] = 0;
        matrixCoeffs[5] = 0;
        AffineTransform linear = new AffineTransform(matrixCoeffs);
        
        Point2D.Double source = new Point2D.Double(0, 1);
        Point2D.Double dest = new Point2D.Double(0, 0);
        
        linear.transform(source, dest);
        
        // return length of the mapped j vector
        return (Math.sqrt(dest.x*dest.x + dest.y*dest.y));
    }
    
    
    
    /**
     * Used to determine if a given style is already in the global list of styles so that we
     * don't create redundant identical styles.
     */    
    public boolean equals(Object other)
    {
        if (other instanceof ODGTextStyle)
        {
            ODGTextStyle otherText = (ODGTextStyle)other;
            if ((font.equals(otherText.font)) && (color.equals(otherText.color)) && (transformScaleFactor == otherText.transformScaleFactor))
                return true;
            else
                return false;
        }
        else
            return false;
    }
    
    
    private String getFontSizeString()
    {
        // Java font size in points; need to convert to units in use
        double fontSize = font.getSize();
        
        // apply any scaling due to Graphics transform
        fontSize = fontSize*transformScaleFactor;
        
        // scale using the Java-pixels-to-units conversion
        String fontSizeString = scaleValueWithUnits(fontSize);
        return fontSizeString;
    }
    
    
    /**
     * Populate XMLNode fields for XML generation
     */
    private void populateXML()
    {
        // done by base class
        //this.setName("style:style");
        
        // add attributes
        this.setAttribute("style:name", styleName);
        this.setAttribute("style:family", "text");
        this.setAttribute("style:parent-style-name", "standard");
        
        
        // add child for style specification
        XMLElement textProperties = new XMLElement();
        textProperties.setNodeName("style:text-properties");
        
        textProperties.setAttribute("style:font-name", ODGFontFaces.mapJavaFontName(font.getName()));
        textProperties.setAttribute("fo:font-size", getFontSizeString());
        textProperties.setAttribute("fo:color", "#" + (Integer.toHexString(color.getRGB()).substring(2)));
        
        if (font.isBold())
            textProperties.setAttribute("fo:font-weight", "bold");
        
        if (font.isItalic())
            textProperties.setAttribute("fo:font-style", "italic");
        
        this.appendChild(textProperties);
        
    }
    
}
