/*
 * JSVG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */



package com.jsevy.jsvg;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;


/**
 * Class for style elements that apply to graphics rendering for things like stroke characteristics and color.
 * @author jsevy
 *
 */
class SVGGraphicsStyle extends SVGStyle
{

    protected boolean stroked;
    protected boolean filled;
    protected BasicStroke stroke;
    protected Color color;
    
    
    /**
     * Create a new style element for graphics operations.
     * 
     * @param graphics  The Java graphics object, used to get stroke and color info
     * @param stroked   True if the object is to be outlined with the current color
     * @param filled    True if the object is to be filled with the current color
     */
    public SVGGraphicsStyle(Graphics2D graphics, boolean stroked, boolean filled)
    {
        // base class gets style name
        
        this.stroked = stroked;
        this.filled = filled;
        this.color = graphics.getColor();
        
        this.stroke = (BasicStroke)graphics.getStroke();
        
        populateXML();
    }
    
    
    /**
     * Used to determine if a given style is already in the global list of styles so that we
     * don't create redundant identical styles.
     */
    public boolean equals(Object other)
    {
        if (other instanceof SVGGraphicsStyle)
        {
            SVGGraphicsStyle otherGraphics = (SVGGraphicsStyle)other;
            if (    (stroked == otherGraphics.stroked) 
                 && (filled == otherGraphics.filled) 
                 && (color.equals(otherGraphics.color)) 
                 && (stroke.equals(otherGraphics.stroke)) 
               )
                return true;
            else
                return false;
        }
        else
            return false;
    }
    
    
    /**
     * Populate XMLNode fields for XML generation
     */
    private void populateXML()
    {
        // done by base class
        //this.setName("style");
        
        // add attributes
        
        //this.setAttribute("name", styleName));
        
        
        // filled?
        if(filled)
        {
            this.setAttribute("fill-opacity", "1");
            
            // set fill color
            this.setAttribute("fill", "#" + (Integer.toHexString(color.getRGB()).substring(2)));

        }
        else
        {
            this.setAttribute("fill-opacity", "0");
        }
        
        // stroked?
        if(stroked)
        {
            this.setAttribute("stroke-opacity", "1");
            
            // set line color
            this.setAttribute("stroke", "#" + (Integer.toHexString(color.getRGB()).substring(2)));
            
            // set line width - no units, oddly
            this.setAttribute("stroke-width", scaleValue(stroke.getLineWidth()));
            
            // set stroke dash pattern, if any
            float[] dashArray = stroke.getDashArray();
            if (dashArray != null)
            {
                String dashArrayString = new String();
                dashArrayString += scaleValue(dashArray[0]);
                for(int i = 1; i < dashArray.length; i++)
                {
                    dashArrayString += " " + scaleValue(dashArray[i]);
                }
                this.setAttribute("stroke-dasharray", dashArrayString);
            }
            
            // set end cap style
            switch (stroke.getEndCap())
            {
                case BasicStroke.CAP_BUTT:
                {
                    this.setAttribute("stroke-linecap", "butt");
                    break;
                }
            
                case BasicStroke.CAP_ROUND:
                {
                    this.setAttribute("stroke-linecap", "round");
                    break;
                }
            
                case BasicStroke.CAP_SQUARE:
                default:
                {
                    this.setAttribute("stroke-linecap", "square");
                    break;
                }
            
            }
            
            // set line join style
            switch (stroke.getLineJoin())
            {
                case BasicStroke.JOIN_BEVEL:
                {
                    this.setAttribute("stroke-linejoin", "bevel");
                    break;
                }
            
                case BasicStroke.JOIN_ROUND:
                {
                    this.setAttribute("stroke-linejoin", "round");
                    break;
                }
            
                case BasicStroke.JOIN_MITER:
                default:
                {
                    this.setAttribute("stroke-linejoin", "miter");
                    break;
                }
            
            }

        }
        else
        {
            this.setAttribute("stroke-opacity", "0");
        }
        
    }
    
    
    
    
}
