/*
 * JSVG Library
 * 
 *   Copyright (C) 2020, Jonathan Sevy <jsevy@jsevy.com>
 *   
 *   Permission is hereby granted, free of charge, to any person obtaining a copy
 *   of this software and associated documentation files (the "Software"), to deal
 *   in the Software without restriction, including without limitation the rights
 *   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *   copies of the Software, and to permit persons to whom the Software is
 *   furnished to do so, subject to the following conditions:
 *   
 *   The above copyright notice and this permission notice shall be included in all
 *   copies or substantial portions of the Software.
 *   
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 *   SOFTWARE.
 * 
 */


package com.jsevy.jsvg;


import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Arc2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;



/**
 * Class representing a path, which is a series of line and curve segments.
 * 
 * @author jsevy
 *
 */
class SVGPath extends SVGEntity
{
    private Shape path;
    
    
    /**
     * Create a new path from a Java Shape object.
     * @param path          The Java Shape defining the path
     * @param style         The ODGStyle object specifying rendering parameters for the shape (color, thickness, etc.)
     * @param transform     Affine transform to be applied to the shape
     */
    public SVGPath(Shape path, SVGGraphicsStyle style, AffineTransform transform)
    {
        super(style, transform);
        
        this.path = path;
        this.style = style;
        
        populateXML();
    }
    
    /**
     * Populate XMLNode fields for XML generation
     */
    protected void populateXML()
    {
        
        this.setNodeName("path");
        
        // add attributes, including the path data
        StringBuffer pathData = new StringBuffer();
        
        // need these to determine location of path/viewport
        double minX = 0;
        double minY = 0;
        double maxX = 0;
        double maxY = 0;
        
        // first, see if the shape is an arc; if so, handle directly without path iterator
        if (path instanceof Arc2D) 
        {
            
            // get parameters
            Arc2D arc = (Arc2D)path;
            
            // start and end points
            Point2D startPoint = arc.getStartPoint();
            Point2D endPoint = arc.getEndPoint();
            
            // x and y radii - come from half width and height of bounding rectangle
            Rectangle boundingRect = arc.getBounds();
            double xRadius = boundingRect.getWidth()/2;
            double yRadius = boundingRect.getHeight()/2;
            
            // this may be mapped to a new angle if we pre-apply the transform
            double xAxisRotationDegrees = 0;
            
            // arc flags - gah...
            //double angleStartDegrees = arc.getAngleStart();
            double angleExtentDegrees = arc.getAngleExtent();
            int largeArcFlag;
            int arcSweepFlag;
            
            if (Math.abs(angleExtentDegrees) > 180)
            {
                largeArcFlag = 1;
            }
            else
            {
                largeArcFlag = 0;
            }
            
            if (angleExtentDegrees < 0)
            {
                arcSweepFlag = 1;
            }
            else
            {
                arcSweepFlag = 0;
            }
            
            // use precise bounding box to get parameters for viewport
            Rectangle2D preciseBoundingRect = arc.getBounds2D();
            
            minX = preciseBoundingRect.getMinX();
            minY = preciseBoundingRect.getMinY();
            maxX = preciseBoundingRect.getMaxX();
            maxY = preciseBoundingRect.getMaxY();
            
            
            
            // moveto start point as initial operation in path data
            pathData.append("M " + scaleValue(startPoint.getX()) + " " + scaleValue(startPoint.getY()) + " ");
            
            // append data for arc
            pathData.append("A " + scaleValue(xRadius) + " " + scaleValue(yRadius) + " " 
                                 + scaleValue(xAxisRotationDegrees) + " " + largeArcFlag + " " + arcSweepFlag + " "
                                 + scaleValue(endPoint.getX()) + " " + scaleValue(endPoint.getY()) + " ");
            
            // we need to close appropriately
            switch (arc.getArcType())
            {
                case Arc2D.CHORD:
                {
                    // join end points - just add "Z"
                    pathData.append("Z" + " ");
                    break;
                }
                
                case Arc2D.PIE:
                {
                    // add line from endpoint to center, then close
                    pathData.append("L " + scaleValue(endPoint.getX()) + " " + scaleValue(endPoint.getY()) + " " + scaleValue(arc.getCenterX()) + " " + scaleValue(arc.getCenterY()) + " " + "Z" + " ");
                    break;
                }
                
                case Arc2D.OPEN:
                default:
                {
                    // nothing to add here
                    break;
                }
            }
            
        }
        else
        {
            // add path data, using a PathIterator; note that path data doesn't want units...
            PathIterator iterator = path.getPathIterator(new AffineTransform());
            
            
                        
            double[] coords = new double[6]; 
            int curveType = iterator.currentSegment(coords);
            
            // first operation must be a moveto; if not we don't put in any path data
            // we also use the coordinates to find the x and y location of the viewport
            curveType = iterator.currentSegment(coords);
            if (curveType == PathIterator.SEG_MOVETO)
            {
                
                double currentX = coords[0];
                double currentY = coords[1];
                
                // for "Z" operations
                double startX = coords[0];
                double startY = coords[1];
                
                minX = coords[0];
                minY = coords[1];
                maxX = coords[0];
                maxY = coords[1];
                
                // add as initial operation in path data for use in subsequent SEG_CLOSE operation
                pathData.append("M " + scaleValue(coords[0]) + " " + scaleValue(coords[1]) + " ");
                
                iterator.next();
                
                // now iterate through remaining path segments, adding path data for each
                while (!iterator.isDone())
                {
                   curveType = iterator.currentSegment(coords);
                    
                    switch (curveType)
                    {
                        case PathIterator.SEG_MOVETO:
                        {
                            // set as initial point for use in subsequent SEG_CLOSE operation
                            pathData.append("M " + scaleValue(coords[0]) + " " + scaleValue(coords[1]) + " ");
                            
                            startX = coords[0];
                            startY = coords[1];
                            
                            // check for minX and minY for viewport coords
                            minX = Math.min(minX,coords[0]);
                            minY = Math.min(minY,coords[1]);
                            maxX = Math.max(maxX,coords[0]);
                            maxY = Math.max(maxY,coords[1]);
                            
                            currentX = coords[0];
                            currentY = coords[1];
                            
                            break;
                        }
                        
                        case PathIterator.SEG_LINETO:
                        {
                            pathData.append("L " + scaleValue(coords[0]) + " " + scaleValue(coords[1]) + " ");
                            
                            // check for minX and minY for viewport coords
                            minX = Math.min(minX,coords[0]);
                            minY = Math.min(minY,coords[1]);
                            maxX = Math.max(maxX,coords[0]);
                            maxY = Math.max(maxY,coords[1]);
                            
                            currentX = coords[0];
                            currentY = coords[1];
                            
                            break;
                        }
                        
                        case PathIterator.SEG_QUADTO:
                        {
                            pathData.append("Q " + scaleValue(coords[0]) + " " + scaleValue(coords[1]) + " " 
                                                 + scaleValue(coords[2]) + " " + scaleValue(coords[3]) + " ");
                            
                            // check for minX and minY for viewport coords
                            MinMax minMaxX = findQuadraticBezierMaxMin(currentX, coords[0], coords[2]);
                            MinMax minMaxY = findQuadraticBezierMaxMin(currentY, coords[1], coords[3]);
                            minX = Math.min(minX, minMaxX.min);
                            minY = Math.min(minY, minMaxY.min);
                            maxX = Math.max(maxX, minMaxX.max);
                            maxY = Math.max(maxY, minMaxY.max);
                            
                            currentX = coords[2];
                            currentY = coords[3];
                            
                            break;
                        }
                        
                        case PathIterator.SEG_CUBICTO:
                        {
                            pathData.append("C " + scaleValue(coords[0]) + " " + scaleValue(coords[1]) + " " 
                                                 + scaleValue(coords[2]) + " " + scaleValue(coords[3]) + " " 
                                                 + scaleValue(coords[4]) + " " + scaleValue(coords[5]) + " ");
                            
                            // check for minX and minY for viewport coords
                            MinMax minMaxX = findCubicBezierMinMax(currentX, coords[0], coords[2], coords[4]);
                            MinMax minMaxY = findCubicBezierMinMax(currentY, coords[1], coords[3], coords[5]);
                            minX = Math.min(minX, minMaxX.min);
                            minY = Math.min(minY, minMaxY.min);
                            maxX = Math.max(maxX, minMaxX.max);
                            maxY = Math.max(maxY, minMaxY.max);
                            
                            currentX = coords[4];
                            currentY = coords[5];
                            
                            break;
                        }
                        
                        case PathIterator.SEG_CLOSE:
                        {
                            pathData.append("Z ");
                            
                            // set current position to last moveto position
                            currentX = startX;
                            currentY = startY;
                            
                            break;
                        }
                        
                        default:
                        {
                            // do nothing
                            break;
                        }
                    }
                    
                    iterator.next();
                    
                }
            }
        }
        
        // set location of viewport containing the path in x and y attributes
        this.setAttribute("x", scaleValue(minX));
        this.setAttribute("y", scaleValue(minY));
        
        // need width and height and viewbox to set scale
        this.setAttribute("width", scaleValue(Math.abs(maxX)));
        this.setAttribute("height", scaleValue(Math.abs(maxY)));
        this.setAttribute("viewBox", "0 0 " + scaleValue(Math.abs(maxX)) + " " + scaleValue(Math.abs(maxY)));
        
        this.setAttribute("d", pathData.toString());
        
        // add style attributes, and transform attribute if any
        super.populateXML();
        
    }
    
    
    private class MinMax
    {
        public double min = Double.MAX_VALUE;
        public double max = Double.MIN_VALUE;
    }
    
    private MinMax findQuadraticBezierMaxMin(double p0, double p1, double p2)
    {
        MinMax minMax = new MinMax();
        
        // first, check endpoints
        minMax.min = Math.min(p0, p2);
        minMax.max = Math.max(p0, p2);
        
        // then, check extremum, unless degenerate linear
        if ((p0 + p2 - 2*p1) != 0)
        {
            double t = (p0 - p1)/(p0 + p2 - 2*p1);
            
            // make sure it lies within the curve
            if ((t > 0) && (t < 1))
            {
                double value = quadraticBezierValue(p0, p1, p2, t);
                minMax.min = Math.min(minMax.min, value);
                minMax.max = Math.max(minMax.max, value);
            }
        }
        
        return minMax;
    }
    
    
    private double quadraticBezierValue(double p0, double p1, double p2, double t)
    {
        return p0*(1-t)*(1-t) + 2*p1*(1-t)*t + p2*t*t;
    }
    
    
    
    private MinMax findCubicBezierMinMax(double p0, double p1, double p2, double p3)
    {
        MinMax minMax = new MinMax();
        
        // first, check endpoints
        minMax.min = Math.min(p0, p3);
        minMax.max = Math.max(p0, p3);
        
        
        // then, check extremum, unless degenerate linear
        // solving a quadratic
        //double a = p1 + p3 - p0 - 3*p2;
        //double b = 2*(p0 - p1 + p2);
        //double c = p1 - p0;
        double a = 3*p1 - 3*p2 + p3 - p0;
        double b = 2*p0 - 4*p1 + 2*p2;
        double c = p1 - p0;
        
        if (a != 0)
        {
            double root = Math.sqrt(b*b - 4*a*c);
            double tPlus = (-b + root)/(2*a);
            double tMinus = (-b - root)/(2*a);
            
            // make sure they lie within the curve
            if ((tPlus > 0) && (tPlus < 1))
            {
                double value = cubicBezierValue(p0, p1, p2, p3, tPlus);
                minMax.min = Math.min(minMax.min, value);
                minMax.max = Math.min(minMax.max, value);
            }
            
            if ((tMinus > 0) && (tMinus < 1))
            {
                double value = cubicBezierValue(p0, p1, p2, p3, tMinus);
                minMax.min = Math.min(minMax.min, value);
                minMax.max = Math.min(minMax.max, value);
            }
        }
        
        return minMax;
    }
    
    
    private double cubicBezierValue(double p0, double p1, double p2, double p3, double t)
    {
        return p0*(1-t)*(1-t)*(1-t) + 3*p1*(1-t)*(1-t)*t + 3*p2*(1-t)*t*t + p3*t*t*t;
    }
    
    
    
}