package com.jsevy.jxml;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.TypeInfo;

public class XMLElement extends XMLNode implements Element
{
    
    // only elements have attributes
    protected XMLNamedNodeMap attributes;
    
    
    
    
    public XMLElement(String name, String value)
    {
        super(name, value);
        
        this.attributes = new XMLNamedNodeMap();
    }
    
    
    public XMLElement(String name)
    {
        this(name, "");
    }
    
    
    public XMLElement()
    {
        this("");
    }
    
    
    @Override
    public Node cloneNode(boolean deep)
    {
        // the base-class Node method takes care of kids, but not attributes since only elements have them
        XMLElement newElement = (XMLElement) super.cloneNode(deep);
        
        // clone attributes
        newElement.attributes = new XMLNamedNodeMap();
        for (int i = 0; i < attributes.getLength(); i++)
        {
            XMLAttribute attribute = (XMLAttribute)attributes.item(i);
            newElement.setAttributeNode((Attr)attribute.cloneNode(true));
        }
        
        return newElement;
    }
    
    
    public void setOwnerDocument(Document owner, boolean deep)
    {
        super.setOwnerDocument(owner, deep);
        
        // do attributes recursively if deep and possible...
        if (deep)
        {
            for (int i = 0; i < attributes.getLength(); i++)
            {
                Attr attribute = (Attr)attributes.item(i);
                if (attribute instanceof XMLAttribute)
                    ((XMLAttribute)attribute).setOwnerDocument(owner, deep);
            }
        }
    }
    
    
    @Override
    public short compareDocumentPosition(Node other) throws DOMException
    {
        // first check if this is one of our attributes
        if (other.isSameNode(attributes.getNamedItem(other.getNodeName())))
        {
            // other is one of our attributes - by spec, it is contained by us and follows us
            return (Node.DOCUMENT_POSITION_CONTAINED_BY | Node.DOCUMENT_POSITION_FOLLOWING);
        }
        
        // otherwise check children, etc.
        return super.compareDocumentPosition(other);
    }
    
    
    @Override
    public short getNodeType()
    {
        return Node.ELEMENT_NODE;
    }


    @Override
    public String getTagName()
    {
        // will be the full qualified name if there's a prefix
        return this.getNodeName();
    }
    
    
    @Override
    public NamedNodeMap getAttributes()
    {
        return attributes;    
    }


    @Override
    public String getAttribute(String name)
    {
        XMLAttribute attribute = (XMLAttribute)attributes.getNamedItem(name);
        if (attribute != null)
            return attribute.getValue();
        else
            return "";
        
    }


    @Override
    public void setAttribute(String name, String value) throws DOMException
    {
        XMLAttribute attribute = new XMLAttribute(name, value);
        attribute.setOwnerElement(this);
        attributes.setNamedItem(attribute);
    }


    @Override
    public void removeAttribute(String name) throws DOMException
    {
        attributes.removeNamedItem(name);
    }


    @Override
    public Attr getAttributeNode(String name)
    {
        return (Attr)attributes.getNamedItem(name);
    }


    @Override
    public Attr setAttributeNode(Attr newAttr) throws DOMException
    {
        // this is annoying - the XML spec for Attr has the ownerElement member readonly, and
        // doesn't provide a setter method. And no Attr creation method has a way to set this. 
        // So we can only set this if we have an XMLAttribute instance, or we clone the damn thing
        // into an XMLAttribute. Grrr.... Not a big fan of whoever created the XML spec...
        
        XMLAttribute attribute;
        if (newAttr instanceof XMLAttribute)
        {
            attribute = (XMLAttribute)newAttr;
        }
        else
        {
            attribute = new XMLAttribute(newAttr.getName(), newAttr.getValue());
            
            // could have children... sigh... clone kids recursively
            for (int i = 0; i < newAttr.getChildNodes().getLength(); i++)
            {
                Node child = newAttr.getChildNodes().item(i);
                Node newChild = child.cloneNode(true);
                attribute.appendChild(newChild);
            }
        }
            
        // now can set owner element appropriately... gah...
        attribute.setOwnerElement(this);
        
        return (Attr)attributes.setNamedItem(attribute);
    }


    @Override
    public Attr removeAttributeNode(Attr oldAttr) throws DOMException
    {
        return (Attr)attributes.removeNamedItem(oldAttr.getName());
    }
    
    
    public void setAttributeNodes(NamedNodeMap attrList)
    {
        for (int i = 0; i < attrList.getLength(); i++)
        {
            // create clones
            Attr attribute = (Attr)attrList.item(i);
            this.setAttributeNode((Attr)attribute.cloneNode(true));
        }
    }


    @Override
    public NodeList getElementsByTagName(String tagName)
    {
        // return all children which are elements with the specified tag name
        XMLNodeList elementList = new XMLNodeList();
        
        for (int i = 0; i < this.getChildNodes().getLength(); i++)
        {
            Node child = this.getChildNodes().item(i);
            if ((child.getNodeType() == Node.ELEMENT_NODE))
            {
                if ((tagName.equals("*")) || (child.getNodeName().equals(tagName)))
                    elementList.add(child);
            
                // now recurse
                elementList.addAll((XMLNodeList)((Element)child).getElementsByTagName(tagName));
            }
        }
        
        return elementList;
    }


    @Override
    public String getAttributeNS(String namespaceURI, String localName) throws DOMException
    {
        
        Node attribute = attributes.getNamedItemNS(namespaceURI, localName);
        if (attribute != null)
        {
            return attribute.getNodeValue();
        }
        else
        {
            return "";
        }
    }


    @Override
    public void setAttributeNS(String namespaceURI, String qualifiedName, String value) throws DOMException
    {
        XMLAttribute attribute = new XMLAttribute(qualifiedName, value);
        attribute.setNamespaceURI(namespaceURI);
        attributes.setNamedItemNS(attribute);
    }


    @Override
    public void removeAttributeNS(String namespaceURI, String localName) throws DOMException
    {
        attributes.removeNamedItemNS(namespaceURI, localName);
    }


    @Override
    public Attr getAttributeNodeNS(String namespaceURI, String localName) throws DOMException
    {
        return (Attr)attributes.getNamedItemNS(namespaceURI, localName);
    }


    @Override
    public Attr setAttributeNodeNS(Attr newAttr) throws DOMException
    {
        return (Attr)attributes.setNamedItemNS(newAttr);
    }


    @Override
    public NodeList getElementsByTagNameNS(String namespaceURI, String localName) throws DOMException
    {
        // return all children which are elements with the specified tag name and namespaceURI
        XMLNodeList elementList = new XMLNodeList();
        
        for (int i = 0; i < this.getChildNodes().getLength(); i++)
        {
            Node child = this.getChildNodes().item(i);
            if (child.getNodeType() == Node.ELEMENT_NODE)
            {
                if (
                        ((namespaceURI.equals("*")) || (child.getNamespaceURI().equals(namespaceURI)))
                     && 
                        ((localName.equals("*") || (child.getLocalName().equals(localName))))
                   )
                    elementList.add(child);
            
                // now recurse
                elementList.addAll((XMLNodeList)((Element)child).getElementsByTagNameNS(namespaceURI, localName));
            }
        }
        
        return elementList;
    }
    
    
    @Override
    public boolean hasAttributes()
    {
        return (attributes.size() > 0);
    }


    @Override
    public boolean hasAttribute(String name)
    {
        Attr attribute = this.getAttributeNode(name);
        if (attribute != null)
            return true;
        else
            return false;
    }


    @Override
    public boolean hasAttributeNS(String namespaceURI, String localName) throws DOMException
    {
        Attr attribute = this.getAttributeNodeNS(namespaceURI, localName);
        if (attribute != null)
            return true;
        else
            return false;
    }


    @Override
    public TypeInfo getSchemaTypeInfo()
    {
        // TODO Auto-generated method stub
        return null;
    }


    @Override
    public void setIdAttribute(String name, boolean isId) throws DOMException
    {
        XMLAttribute attribute = (XMLAttribute)this.getAttributeNode(name);
        if (attribute == null)
            throw new DOMException(DOMException.NOT_FOUND_ERR, "No attribute with name " + name);
        
        attribute.setId(isId);
    }


    @Override
    public void setIdAttributeNS(String namespaceURI, String localName, boolean isId) throws DOMException
    {
        XMLAttribute attribute = (XMLAttribute)this.getAttributeNodeNS(namespaceURI, localName);
        if (attribute == null)
            throw new DOMException(DOMException.NOT_FOUND_ERR, "No attribute with name " + localName);
        
        attribute.setId(isId);
    }


    @Override
    public void setIdAttributeNode(Attr idAttr, boolean isId) throws DOMException
    {
        XMLAttribute attribute = (XMLAttribute)this.getAttributeNode(idAttr.getName());
        if (attribute == null)
            throw new DOMException(DOMException.NOT_FOUND_ERR, "No attribute with name " + idAttr.getName());
        
        attribute.setId(isId);
    }
    
    
    
    @Override
    public String getTextContent() throws DOMException
    {
        // return depends on subclass, so overridden
        // for Element node, return concatenation of the textContent attribute value of every child node,
        // excluding COMMENT_NODE and PROCESSING_INSTRUCTION_NODE nodes. This is the empty string if the 
        // node has no children.
        
        StringBuffer text = new StringBuffer();
        
        for (int i = 0; i < this.getChildNodes().getLength(); i++)
        {
            // recursively get text from nodes except comment nodes and processing instruction nodes
            Node child = this.getChildNodes().item(i);
            short nodeType = child.getNodeType();
            if ((nodeType != Node.COMMENT_NODE) && (nodeType != Node.PROCESSING_INSTRUCTION_NODE))
            {
                String childText = child.getTextContent();
                
                // filter out null results since StringBuffer appends the string "null"... :-\
                if (childText != null)
                    text.append(childText);
            }
        }
        
        return text.toString();
    }
    
}
